import project1.project1

/-
The ultimate goal here is to work toward's Fodor's Lemma, which states that
if S is a stationary subset of a regular cardinal κ and f : S → κ is
regressive (meaning f(α) < α for all α ∈ S), then there exists a unique β
such that f⁻¹({β}) is stationary in κ.

A set is stationary if it intersects every club set somewhere (in some sense 
this is a notion of "largeness")

There is a bunch of scuffed stuff here in the meantime that I need to prove 
this (namely that larger intersections of club sets are club and diagonal
intersections of club sets are club). I've gotten started, but still have a
ways to go.
-/


universe u
variables (S : set ordinal.{u}) (α η κ : ordinal.{u})
(f : ordinal.{u} → set ordinal.{u})

/-
If κ is regular and uncountable, an intersection of <κ-many clubs is club.
I'm going to treat this result as a black box for the purposes of my project
-/
theorem inter_clubs_is_club {κ η : ordinal.{u}}
(f : ordinal.{u} → set ordinal.{u}) (ηltκ : η < κ)
(hclub : ∀ α, α < η → set.is_club (f α) κ) (κreg : κ.cof = κ.card)
(κunctbl : κ.card > cardinal.aleph_0): set.is_club (⋂ (α < η), (f α)) κ :=
begin
  sorry,
end

-- Define the diagonal intersection
def diag_inter := λ (f : ordinal.{u} → set ordinal.{u}),
{β : ordinal.{u} | β < κ ∧ (∀ α < β, β ∈ f α)}

-- And a helpful auxiliary sequence
def D_seq := λ (f : ordinal.{u} → set ordinal.{u}) β, ⋂ (α < β), f α

-- Relate the diagonal intersection to this sequence
lemma diag_inter_D_seq (f : ordinal.{u} → set ordinal.{u}) :
(diag_inter κ f) = {β : ordinal.{u} | β < κ ∧ β ∈ D_seq f β} :=
begin
  ext β,
  simp only [diag_inter, D_seq, set.mem_set_of_eq],
  split,
  { rintro ⟨h1, h2⟩,
    split, exact h1,
    simp [set.mem_Inter_of_mem],
    intros i hi,
    specialize h2 i,
    apply h2 hi, },
  rintro ⟨h1, h2⟩,
  split, exact h1,
  simp [set.mem_Inter_of_mem] at h2,
  exact h2,
end

-- Show that for all α, D α is club
lemma D_seq_club {κ β : ordinal.{u}} (f : ordinal.{u} → set ordinal.{u})
(βltκ : β < κ) (hclub : ∀ α, α < β → set.is_club (f α) κ)
(κreg : κ.cof = κ.card) (κunctbl : κ.card > cardinal.aleph_0) :
set.is_club (D_seq f β) κ :=
begin
  simp only [D_seq],
  exact inter_clubs_is_club f βltκ hclub κreg κunctbl,
end

-- Show the D-sequence is decreasing
lemma D_seq_decreasing {f : ordinal.{u} → set ordinal.{u}}
{β γ : ordinal.{u}} : β ≤ γ → (D_seq f γ) ⊆ (D_seq f β) :=
begin
  intros h x hx,
  dsimp [D_seq] at ⊢ hx,
  simp [set.mem_Inter_of_mem] at ⊢ hx,
  intros i hi,
  specialize hx i,
  apply hx,
  apply lt_of_lt_of_le hi h,
end

-- Show the diagonal intersection of subsets of κ is a subset of κ
lemma diag_inter_subsets_is_subset {f : ordinal.{u} → set ordinal.{u}} : 
set.is_subset_of (diag_inter κ f) κ := 
begin
  rw diag_inter_D_seq,
  dsimp [set.is_subset_of],
  intros β hβ,
  simp only [set.mem_set_of_eq] at hβ,
  exact hβ.1,
end

-- Show the diagonal intersection of closed sets is closed
lemma diag_inter_closed_is_closed {κ : ordinal.{u}}
(f : ordinal.{u} → set ordinal.{u})
(hclosed : ∀ α, α < κ → set.is_closed (f α) κ) : 
set.is_closed (diag_inter κ f) κ :=
begin
  intros A hA Ane supA,
  have : ∀ α < Sup A, Sup A ∈ f α,
  { intros α hα,
    have αltκ : α < κ,
    { apply lt_trans hα supA }, 
    specialize hclosed α αltκ,
    simp [set.is_closed] at hclosed,
    have eβ : ∃ β ∈ A, α < β,
    { exact exists_lt_of_lt_cSup' hα, },
    rcases eβ with ⟨β, βA, hβ⟩,
    have aux1 : β ∈ (D_seq f β),
    { have := set.mem_of_mem_of_subset βA hA,
      rw diag_inter_D_seq at this,
      exact this.2, },
    have aux3 : (D_seq f β) ⊆ f α,
    { dsimp [D_seq],
      apply set.bInter_subset_of_mem,
      exact hβ, },
    have aux4 := set.mem_of_mem_of_subset aux1 aux3,
    have aux2 : ∀ γ ∈ A, β ≤ γ → γ ∈ (D_seq f γ),
    { intros γ γA hγ,
      have := set.mem_of_mem_of_subset γA hA,
      rw diag_inter_D_seq at this,
      exact this.2, },
    have aux5 : ∀ γ ∈ A, β ≤ γ → γ ∈ f α,
    { intros γ γA hγ,
      specialize aux2 γ γA hγ,
      apply set.mem_of_mem_of_subset aux2 _,
      dsimp [D_seq],
      apply set.bInter_subset_of_mem,
      exact lt_of_lt_of_le hβ hγ,
    },
    have s1 : {γ ∈ A | β ≤ γ} ⊆ A,
    { intros γ hγ,
      simp only [set.mem_sep_eq] at hγ,
      exact hγ.1},
    have bdd : bdd_above A,
    { use κ,
      intros a ha,
      have : set.is_subset_of (diag_inter κ f) κ,
      { apply diag_inter_subsets_is_subset, },
        apply le_of_lt,
        apply this.lt _ _,
        apply set.mem_of_mem_of_subset ha hA, },
    have tail_bdd : bdd_above {γ ∈ A | β ≤ γ},
    { use κ,
      intros a ha,
      have : set.is_subset_of (diag_inter κ f) κ,
      { apply diag_inter_subsets_is_subset, },
        apply le_of_lt,
        apply this.lt _ _,
        simp [set.mem_sep_eq] at ha,
        apply set.mem_of_mem_of_subset ha.1 hA, },
    have tail : Sup A = Sup {γ ∈ A | β ≤ γ},
    { apply le_antisymm,
      { apply cSup_le,
        exact Ane,
        intros b hb,
        by_cases b ≤ β,
        { apply le_trans h,
          apply le_cSup tail_bdd,
          simp only [set.mem_sep_eq],
          split,
          exact βA,
          exact le_refl β, },
        apply le_cSup tail_bdd,
        simp only [set.mem_sep_eq],
        split,
        exact hb,
        exact le_of_not_ge h, },
      apply cSup_le_cSup bdd,
      { use β,
        simp only [set.mem_sep_eq],
        split,
        exact βA,
        exact le_refl β, },
      exact s1, },
    rw tail,
    apply hclosed,
    { 
      intros γ hγ,
      simp only [set.mem_sep_eq] at hγ,
      specialize aux5 γ,
      apply aux5 hγ.1 hγ.2, },
    { use β,
      simp only [set.mem_sep_eq],
      split,
      exact βA,
      exact le_refl β, },
    rw ←tail,
    
    exact supA },
  dsimp [diag_inter],
  split,
  exact supA,
  exact this,
end

-- Existence lemma for unbounded sequence
lemma tail_D_seq_nonempty {κ β γ : ordinal.{u}}
(f : ordinal.{u} → set ordinal.{u}) (hclub : ∀ α, α < κ → set.is_club (f α) κ)
(βltκ : β < κ) (γltκ : γ < κ) (κreg : κ.cof = κ.card)
(κunctbl : κ.card > cardinal.aleph_0) : ∃ δ : ordinal.{u},
δ ∈ (D_seq f γ) ∧ δ > β ∧ δ < κ :=
begin
  have aux1 : set.is_club (D_seq f γ) κ,
  { have : ∀ (α : ordinal), α < γ → (f α).is_club κ,
    { intros α αltβ,
      apply hclub,
      apply lt_trans,
      exact αltβ,
      exact γltκ },
    exact D_seq_club f γltκ this κreg κunctbl, },
  have aux2 := aux1.is_unbounded (D_seq f γ) κ,
  dsimp [set.is_unbounded] at aux2,
  specialize aux2 β βltκ,
  rcases aux2 with ⟨δ, ⟨δinDγ, βltδ⟩⟩,
  use δ,
  split,
  exact δinDγ,
  split,
  exact βltδ,
  have aux3 := aux1.is_subset_of (D_seq f γ) κ,
  exact aux3 δinDγ,
end

open set ordinal

-- Construct the sequence used for proof of unboundedness
noncomputable def unbdd_seq {κ β : ordinal.{u}}
(f : ordinal.{u} → set ordinal.{u}) (hclub : ∀ α, α < κ → set.is_club (f α) κ) 
(βltκ : β < κ) (κreg : κ.cof = κ.card) (κunctbl : κ.card > cardinal.aleph_0) : 
ℕ → {x : ordinal.{u} // x < κ}
| 0 := ⟨β, βltκ⟩
| (n+1) := let existence := tail_D_seq_nonempty f hclub (unbdd_seq n).prop 
(unbdd_seq n).prop κreg κunctbl in
⟨existence.some, existence.some_spec.right.right⟩

-- Define the supremum of the unbounded sequence
noncomputable def unbdd_sup {κ β : ordinal.{u}}
(f : ordinal.{u} → set ordinal.{u}) (hclub : ∀ α, α < κ → set.is_club (f α) κ) 
(βltκ : β < κ) (κreg : κ.cof = κ.card) (κunctbl : κ.card > cardinal.aleph_0) := 
sup (λ n, (unbdd_seq f hclub βltκ κreg κunctbl n).val)

-- Convenience lemma to actually use the definition of the sequence
lemma unbdd_seq_aux {κ β : ordinal.{u}} (n : ℕ)
(f : ordinal.{u} → set ordinal.{u}) (hclub : ∀ α, α < κ → set.is_club (f α) κ) 
(βltκ : β < κ) (κreg : κ.cof = κ.card) (κunctbl : κ.card > cardinal.aleph_0) : 
unbdd_seq f hclub βltκ κreg κunctbl (n+1) =
let existence := tail_D_seq_nonempty f hclub
(unbdd_seq f hclub βltκ κreg κunctbl n).prop
(unbdd_seq f hclub βltκ κreg κunctbl n).prop κreg κunctbl in
⟨existence.some, existence.some_spec.right.right⟩ :=
begin
  simp only [unbdd_seq, gt_iff_lt],
  congr',
end

-- First show single steps of the sequence are strictly increasing
lemma unbdd_seq_increasing_step {κ β : ordinal.{u}} {n : ℕ}
(f : ordinal.{u} → set ordinal.{u}) (hclub : ∀ α, α < κ → set.is_club (f α) κ) 
(βltκ : β < κ) (κreg : κ.cof = κ.card) (κunctbl : κ.card > cardinal.aleph_0) : 
(unbdd_seq f hclub βltκ κreg κunctbl n).val < (unbdd_seq f hclub βltκ κreg 
κunctbl (n+1)).val :=
begin
  rw unbdd_seq_aux n f hclub βltκ κreg κunctbl,
  let existence := tail_D_seq_nonempty f hclub
  (unbdd_seq f hclub βltκ κreg κunctbl n).prop
  (unbdd_seq f hclub βltκ κreg κunctbl n).prop κreg κunctbl,
  exact existence.some_spec.right.left,
end

-- Show the sequence is strictly increasing
lemma unbdd_seq_increasing {κ β : ordinal.{u}} {m n : ℕ} (mltn : m < n)
(f : ordinal.{u} → set ordinal.{u}) (hclub : ∀ α, α < κ → set.is_club (f α) κ)
(βltκ : β < κ) (κreg : κ.cof = κ.card) (κunctbl : κ.card > cardinal.aleph_0) : 
(unbdd_seq f hclub βltκ κreg κunctbl m).val <
  (unbdd_seq f hclub βltκ κreg κunctbl n).val :=
begin
  induction n with n ih,
  exfalso,
  exact nat.not_lt_zero m mltn,
  rw [nat.succ_eq_add_one] at ⊢ mltn,
  by_cases m=n,
  { rw h,
    exact unbdd_seq_increasing_step f hclub βltκ κreg κunctbl },
  apply lt_trans,
  apply ih,
  exact array.push_back_idx mltn h,
  exact unbdd_seq_increasing_step f hclub βltκ κreg κunctbl,
end

-- Show the supremum is smaller than κ
lemma unbdd_sup_lt_κ {κ β : ordinal.{u}} (f : ordinal.{u} → set ordinal.{u}) 
(hclub : ∀ α, α < κ → set.is_club (f α) κ) (βltκ : β < κ)
(κreg : κ.cof = κ.card) (κunctbl : κ.card > cardinal.aleph_0) :
(unbdd_sup f hclub βltκ κreg κunctbl) < κ :=
begin
  dsimp [unbdd_sup],
  apply @ordinal.sup_lt_ord_lift _ (λ n,
    (unbdd_seq f hclub βltκ κreg κunctbl n).val) κ,
  rw κreg,
  exact κunctbl,
  intro i,
  simp only,
  exact (unbdd_seq f hclub βltκ κreg κunctbl i).property,
end

-- Show the supremum is greater than β
lemma β_lt_unbdd_sup {κ β : ordinal.{u}} (f : ordinal.{u} → set ordinal.{u}) 
(hclub : ∀ α, α < κ → set.is_club (f α) κ) (βltκ : β < κ)
(κreg : κ.cof = κ.card) (κunctbl : κ.card > cardinal.aleph_0) :
β < (unbdd_sup f hclub βltκ κreg κunctbl) :=
begin
  have aux1 : (unbdd_seq f hclub βltκ κreg κunctbl 1).val ≤
    unbdd_sup f hclub βltκ κreg κunctbl,
  { dsimp [unbdd_sup],
    exact ordinal.le_sup (λ n, (unbdd_seq f hclub βltκ κreg κunctbl n).val) 1 },
  have aux2 : 0 < 1 := by norm_num,
  have aux3 := unbdd_seq_increasing aux2 f hclub βltκ κreg κunctbl,
  apply lt_of_lt_of_le aux3 aux1,
end

-- Show the supremum is in the diagonal intersection
lemma unbdd_sup_in_diag_inter {κ β : ordinal.{u}}
(f : ordinal.{u} → set ordinal.{u}) (hclub : ∀ α, α < κ → set.is_club (f α) κ)
(βltκ : β < κ) (κreg : κ.cof = κ.card) (κunctbl : κ.card > cardinal.aleph_0) : 
unbdd_sup f hclub βltκ κreg κunctbl ∈ diag_inter κ f :=
begin
  set seq := λ n, ↑(unbdd_seq f hclub βltκ κreg κunctbl n) with seq_def,
  rw diag_inter_D_seq,
  dsimp,
  split,
  { exact unbdd_sup_lt_κ f hclub βltκ κreg κunctbl },
  simp only [D_seq, set.mem_Inter],
  intros α hα,
  have aux1 : ∃ m : ℕ, α < seq m,
  { refine lt_sup.mp _,
    dsimp [unbdd_sup] at hα,
    simp only,
    rw seq_def,
    exact hα },
  cases aux1 with m hm,
  set seq2 := λ n : ℕ,
  (unbdd_seq f hclub βltκ κreg κunctbl (n+m+1)).val with seq2_def,
  have aux2 : ∀ n ≥ m, seq (n+1) ∈ f α,
  { intros n hn,
    have aux21 : seq (n+1) ∈ D_seq f (seq n),
    { rw seq_def,
      simp only,
      rw unbdd_seq_aux n f hclub βltκ κreg κunctbl,
      let existence := tail_D_seq_nonempty f hclub
      (unbdd_seq f hclub βltκ κreg κunctbl n).prop
      (unbdd_seq f hclub βltκ κreg κunctbl n).prop κreg κunctbl,
      exact existence.some_spec.left },
    have aux22 : D_seq f (seq n) ⊆ D_seq f (seq m),
    { apply D_seq_decreasing,
      rw [seq_def],
      apply le_of_lt_or_eq,
      by_cases m=n,
      right, rw h,
      left,
      apply unbdd_seq_increasing,
      exact lt_of_le_of_ne hn h, },
    have aux23 : D_seq f (seq m) ⊆ f α,
    { dsimp [D_seq],
      apply set.bInter_subset_of_mem,
      exact hm, },
    apply aux23,
    apply aux22,
    exact aux21, },
  dsimp [unbdd_sup],
  have aux3 : sup seq = sup seq2,
  { rw [seq_def, seq2_def],
    apply le_antisymm,
    { apply ordinal.sup_le,
      intro n,
      apply le_trans,
      { have : n < n + m + 1, by linarith, --linarith
        apply le_of_lt,
        apply unbdd_seq_increasing this f hclub βltκ κreg κunctbl },
      apply ordinal.le_sup (λ n : ℕ,
      (unbdd_seq f hclub βltκ κreg κunctbl (n+m+1)).val), },
    apply ordinal.sup_le,
    intro n,
    apply ordinal.le_sup (λ n : ℕ,
      (unbdd_seq f hclub βltκ κreg κunctbl (n)).val), },
  rw [←seq_def, aux3, seq2_def],
  have aux4 : set.range seq2 ⊆ f α,
  { intros x hx,
    simp only [gt_iff_lt, mem_range] at hx,
    cases hx with n hn,
    dsimp [seq2] at hn,
    rw ←hn,
    apply aux2,
    linarith, }, --linarith
  have aux5 : α < κ,
  { apply lt_trans hα (unbdd_sup_lt_κ f hclub βltκ κreg κunctbl) },
  have αclub := hclub α aux5,
  apply αclub.is_closed _ _,
  exact aux4,
  { apply set.nonempty_of_mem,
    apply set.mem_range_self 0, },
  rw [←seq2_def, Sup_eq_sup, ←aux3],
  rw [seq_def],
  exact unbdd_sup_lt_κ f hclub βltκ κreg κunctbl,
end

-- Show that the diagonal intersection is unbounded
lemma diag_inter_club_is_unbounded {κ : ordinal.{u}}
(f : ordinal.{u} → set ordinal.{u}) (hclub : ∀ α, α < κ → set.is_club (f α) κ) 
(κreg : κ.cof = κ.card) (κunctbl : κ.card > cardinal.aleph_0) : 
set.is_unbounded (diag_inter κ f) κ :=
begin
  dsimp [set.is_unbounded],
  intros β βltκ,
  use unbdd_sup f hclub βltκ κreg κunctbl,
  split,
  { exact unbdd_sup_in_diag_inter f hclub βltκ κreg κunctbl },
  { exact β_lt_unbdd_sup f hclub βltκ κreg κunctbl },
end

-- Show that the diagonal intersection is club
theorem diag_inter_clubs_is_club {κ : ordinal.{u}} 
(f : ordinal.{u} → set ordinal.{u}) (hclub : ∀ α, α < κ → set.is_club (f α) κ) 
(κreg : κ.cof = κ.card) (κunctbl : κ.card > cardinal.aleph_0) : 
set.is_club (diag_inter κ f) κ :=
begin
  simp only [set.is_club],
  split,
  { exact diag_inter_subsets_is_subset κ },
  split,
  { apply diag_inter_closed_is_closed f,
    intros α hα,
    specialize hclub α hα,
    exact hclub.is_closed _ _},
  exact diag_inter_club_is_unbounded f hclub κreg κunctbl,
end

-- Now, we establish the definitions needed to state and prove Fodor's lemma
def set.is_stationary (S : set ordinal.{u}) (κ : ordinal.{u}) : Prop :=
S.is_subset_of κ ∧ ∀ (C : set ordinal.{u}), C.is_club κ → (S ∩ C).nonempty
def is_regressive_on (g : ordinal.{u} → ordinal.{u}) (S : set ordinal.{u}) : 
Prop := ∀ α ∈ S, g α < α

-- Show that if a set is nonstationary, there is a club disjoint from it
lemma exists_disjoint_club_of_nonstationary {S : set ordinal.{u}}
{κ : ordinal.{u}} (Ssubsetκ : S.is_subset_of κ) (Snstat : ¬(S.is_stationary κ)) 
: ∃ (C : set ordinal.{u}), set.is_club C κ ∧ S ∩ C = ∅ :=
begin
  dsimp [set.is_stationary] at Snstat,
  push_neg at Snstat,
  have := Snstat Ssubsetκ,
  rcases this with ⟨C, Cclub, Cinter⟩,
  use C,
  split,
  exact Cclub,
  rw ←set.not_nonempty_iff_eq_empty,
  exact Cinter,
end

/-
Given a collection of nonstationary sets, define a function that assigns a club 
witnessing nonstationarity of each
-/
def Fodor_club_fun {κ : ordinal.{u}} {h : ordinal.{u} → set ordinal.{u}} 
(hsubset : ∀ α < κ, (h α).is_subset_of κ)
(hnonstat : ∀ α < κ, ¬(h α).is_stationary κ) : ordinal.{u} → set ordinal.{u}
| α := if hyp : α < κ then (exists_disjoint_club_of_nonstationary
  (hsubset α hyp) (hnonstat α hyp)).some
  else 
  ∅ -- this case never arises in practice

lemma Fodor_club_fun_club {κ : ordinal.{u}} (h : ordinal.{u} → set ordinal.{u}) 
(hsubset : ∀ α < κ, (h α).is_subset_of κ)
(hnonstat : ∀ α < κ, ¬(h α).is_stationary κ) {α : ordinal.{u}} (hyp : α < κ) : 
(Fodor_club_fun hsubset hnonstat α).is_club κ :=
begin
  dsimp [Fodor_club_fun],
  split_ifs,
  exact (exists_disjoint_club_of_nonstationary (hsubset α hyp)
  (hnonstat α hyp)).some_spec.left,
end

lemma Fodor_club_fun_disjoint {κ : ordinal.{u}}
(h : ordinal.{u} → set ordinal.{u}) (hsubset : ∀ α < κ, (h α).is_subset_of κ) 
(hnonstat : ∀ α < κ, ¬(h α).is_stationary κ) {α : ordinal.{u}} (hyp : α < κ) : 
(h α) ∩ (Fodor_club_fun hsubset hnonstat α) = ∅ :=
begin
  dsimp [Fodor_club_fun],
  split_ifs,
  exact (exists_disjoint_club_of_nonstationary (hsubset α hyp)
    (hnonstat α hyp)).some_spec.right,
end

-- Prove Fodor's lemma, which states that a regressive function on stationary 
-- domain has at least one point in its image whose preimage is stationary
theorem Fodors_lemma {κ : ordinal.{u}} {S : set ordinal.{u}}
{g : ordinal.{u} → ordinal.{u}} (Sstat : S.is_stationary κ)
(gregr : is_regressive_on g S) (κreg : κ.cof = κ.card)
(κunctbl : κ.card > cardinal.aleph_0)
(gsubset : ∀ α < κ, set.is_subset_of (g ⁻¹' {α}) κ) : ∃ α < κ,
  set.is_stationary (g ⁻¹' {α}) κ :=
begin
  by_contradiction,
  push_neg at h,
  set C := Fodor_club_fun gsubset h with C_def,
  set A := λ α, (g ⁻¹' {α}) with A_def,
  have aux1 : ∀ (α : ordinal), α < κ → (C α).is_club κ,
  { intros α hα,
    exact Fodor_club_fun_club A gsubset h hα, },
  have aux2 := diag_inter_clubs_is_club C aux1 κreg κunctbl,
  --have aux3 : (S ∩ (diag_inter κ C)).nonempty,
  have aux3 : ∃ β, β ∈ (S ∩ (diag_inter κ C)), 
  { exact Sstat.right _ aux2, },
  cases aux3 with β hβ,
  have aux4 : β ∈ A (g β),
  { rw [A_def, set.mem_preimage],
    exact mem_singleton (g β), },
  have aux51 : g β < β,
    { apply gregr,
      exact hβ.left, },
  have aux5 : β ∈ C (g β),
  { rw diag_inter_D_seq at hβ,
    dsimp [D_seq] at hβ,
    apply mem_of_subset_of_mem,
    swap,
    exact hβ.right.right,
    exact Inter₂_subset _ aux51, },
  have aux6 := mem_inter aux4 aux5,
  have aux7 : A (g β) ∩ C (g β) = ∅,
  { apply Fodor_club_fun_disjoint,
    apply lt_trans,
    exact aux51,
    apply Sstat.left,
    exact hβ.left, },
  rw [←set.not_nonempty_iff_eq_empty] at aux7,
  exact aux7 (set.nonempty_of_mem aux6),
end

